@@ -2,6 +2,7 @@ use crate::fs::{OpenOptions, Permissions};
 use crate::fs_utf8::{from_utf8, to_utf8, DirBuilder, File, Metadata, ReadDir};
 #[cfg(unix)]
 use crate::os::unix::net::{UnixDatagram, UnixListener, UnixStream};
+use camino::{Utf8Path, Utf8PathBuf};
 use cap_primitives::{ambient_authority, AmbientAuthority};
 #[cfg(not(windows))]
 use io_lifetimes::{AsFd, BorrowedFd, FromFd, IntoFd, OwnedFd};
@@ -57,7 +58,7 @@ impl Dir {
     /// This corresponds to [`std::fs::File::open`], but only accesses paths
     /// relative to `self`.
     #[inline]
-    pub fn open<P: AsRef<str>>(&self, path: P) -> io::Result<File> {
+    pub fn open<P: AsRef<Utf8Path>>(&self, path: P) -> io::Result<File> {
         let path = from_utf8(path)?;
         self.cap_std.open(path).map(File::from_cap_std)
     }
@@ -69,7 +70,11 @@ impl Dir {
     /// Instead of being a method on `OpenOptions`, this is a method on `Dir`,
     /// and it only accesses paths relative to `self`.
     #[inline]
-    pub fn open_with<P: AsRef<str>>(&self, path: P, options: &OpenOptions) -> io::Result<File> {
+    pub fn open_with<P: AsRef<Utf8Path>>(
+        &self,
+        path: P,
+        options: &OpenOptions,
+    ) -> io::Result<File> {
         let path = from_utf8(path)?;
         self.cap_std
             .open_with(path, options)
@@ -78,7 +83,7 @@ impl Dir {
 
     /// Attempts to open a directory.
     #[inline]
-    pub fn open_dir<P: AsRef<str>>(&self, path: P) -> io::Result<Self> {
+    pub fn open_dir<P: AsRef<Utf8Path>>(&self, path: P) -> io::Result<Self> {
         let path = from_utf8(path)?;
         self.cap_std.open_dir(path).map(Self::from_cap_std)
     }
@@ -88,7 +93,7 @@ impl Dir {
     /// This corresponds to [`std::fs::create_dir`], but only accesses paths
     /// relative to `self`.
     #[inline]
-    pub fn create_dir<P: AsRef<str>>(&self, path: P) -> io::Result<()> {
+    pub fn create_dir<P: AsRef<Utf8Path>>(&self, path: P) -> io::Result<()> {
         let path = from_utf8(path)?;
         self.cap_std.create_dir(path)
     }
@@ -99,7 +104,7 @@ impl Dir {
     /// This corresponds to [`std::fs::create_dir_all`], but only accesses
     /// paths relative to `self`.
     #[inline]
-    pub fn create_dir_all<P: AsRef<str>>(&self, path: P) -> io::Result<()> {
+    pub fn create_dir_all<P: AsRef<Utf8Path>>(&self, path: P) -> io::Result<()> {
         let path = from_utf8(path)?;
         self.cap_std.create_dir_all(path)
     }
@@ -109,7 +114,7 @@ impl Dir {
     ///
     /// This corresponds to [`std::fs::DirBuilder::create`].
     #[inline]
-    pub fn create_dir_with<P: AsRef<str>>(
+    pub fn create_dir_with<P: AsRef<Utf8Path>>(
         &self,
         path: P,
         dir_builder: &DirBuilder,
@@ -123,7 +128,7 @@ impl Dir {
     /// This corresponds to [`std::fs::File::create`], but only accesses paths
     /// relative to `self`.
     #[inline]
-    pub fn create<P: AsRef<str>>(&self, path: P) -> io::Result<File> {
+    pub fn create<P: AsRef<Utf8Path>>(&self, path: P) -> io::Result<File> {
         let path = from_utf8(path)?;
         self.cap_std.create(path).map(File::from_cap_std)
     }
@@ -135,7 +140,7 @@ impl Dir {
     /// an absolute path, returns a path relative to the directory
     /// represented by `self`.
     #[inline]
-    pub fn canonicalize<P: AsRef<str>>(&self, path: P) -> io::Result<String> {
+    pub fn canonicalize<P: AsRef<Utf8Path>>(&self, path: P) -> io::Result<Utf8PathBuf> {
         let path = from_utf8(path)?;
         self.cap_std.canonicalize(path).and_then(to_utf8)
     }
@@ -147,7 +152,7 @@ impl Dir {
     /// This corresponds to [`std::fs::copy`], but only accesses paths
     /// relative to `self`.
     #[inline]
-    pub fn copy<P: AsRef<str>, Q: AsRef<str>>(
+    pub fn copy<P: AsRef<Utf8Path>, Q: AsRef<Utf8Path>>(
         &self,
         from: P,
         to_dir: &Self,
@@ -163,7 +168,7 @@ impl Dir {
     /// This corresponds to [`std::fs::hard_link`], but only accesses paths
     /// relative to `self`.
     #[inline]
-    pub fn hard_link<P: AsRef<str>, Q: AsRef<str>>(
+    pub fn hard_link<P: AsRef<Utf8Path>, Q: AsRef<Utf8Path>>(
         &self,
         src: P,
         dst_dir: &Self,
@@ -180,7 +185,7 @@ impl Dir {
     /// This corresponds to [`std::fs::metadata`], but only accesses paths
     /// relative to `self`.
     #[inline]
-    pub fn metadata<P: AsRef<str>>(&self, path: P) -> io::Result<Metadata> {
+    pub fn metadata<P: AsRef<Utf8Path>>(&self, path: P) -> io::Result<Metadata> {
         let path = from_utf8(path)?;
         self.cap_std.metadata(path)
     }
@@ -196,7 +201,7 @@ impl Dir {
     /// This corresponds to [`std::fs::read_dir`], but only accesses paths
     /// relative to `self`.
     #[inline]
-    pub fn read_dir<P: AsRef<str>>(&self, path: P) -> io::Result<ReadDir> {
+    pub fn read_dir<P: AsRef<Utf8Path>>(&self, path: P) -> io::Result<ReadDir> {
         let path = from_utf8(path)?;
         self.cap_std.read_dir(path).map(ReadDir::from_cap_std)
     }
@@ -206,7 +211,7 @@ impl Dir {
     /// This corresponds to [`std::fs::read`], but only accesses paths
     /// relative to `self`.
     #[inline]
-    pub fn read<P: AsRef<str>>(&self, path: P) -> io::Result<Vec<u8>> {
+    pub fn read<P: AsRef<Utf8Path>>(&self, path: P) -> io::Result<Vec<u8>> {
         let path = from_utf8(path)?;
         self.cap_std.read(path)
     }
@@ -216,7 +221,7 @@ impl Dir {
     /// This corresponds to [`std::fs::read_link`], but only accesses paths
     /// relative to `self`.
     #[inline]
-    pub fn read_link<P: AsRef<str>>(&self, path: P) -> io::Result<String> {
+    pub fn read_link<P: AsRef<Utf8Path>>(&self, path: P) -> io::Result<Utf8PathBuf> {
         let path = from_utf8(path)?;
         self.cap_std.read_link(path).and_then(to_utf8)
     }
@@ -226,7 +231,7 @@ impl Dir {
     /// This corresponds to [`std::fs::read_to_string`], but only accesses
     /// paths relative to `self`.
     #[inline]
-    pub fn read_to_string<P: AsRef<str>>(&self, path: P) -> io::Result<String> {
+    pub fn read_to_string<P: AsRef<Utf8Path>>(&self, path: P) -> io::Result<String> {
         let path = from_utf8(path)?;
         self.cap_std.read_to_string(path)
     }
@@ -236,7 +241,7 @@ impl Dir {
     /// This corresponds to [`std::fs::remove_dir`], but only accesses paths
     /// relative to `self`.
     #[inline]
-    pub fn remove_dir<P: AsRef<str>>(&self, path: P) -> io::Result<()> {
+    pub fn remove_dir<P: AsRef<Utf8Path>>(&self, path: P) -> io::Result<()> {
         let path = from_utf8(path)?;
         self.cap_std.remove_dir(path)
     }
@@ -247,7 +252,7 @@ impl Dir {
     /// This corresponds to [`std::fs::remove_dir_all`], but only accesses
     /// paths relative to `self`.
     #[inline]
-    pub fn remove_dir_all<P: AsRef<str>>(&self, path: P) -> io::Result<()> {
+    pub fn remove_dir_all<P: AsRef<Utf8Path>>(&self, path: P) -> io::Result<()> {
         let path = from_utf8(path)?;
         self.cap_std.remove_dir_all(path)
     }
@@ -278,7 +283,7 @@ impl Dir {
     /// This corresponds to [`std::fs::remove_file`], but only accesses paths
     /// relative to `self`.
     #[inline]
-    pub fn remove_file<P: AsRef<str>>(&self, path: P) -> io::Result<()> {
+    pub fn remove_file<P: AsRef<Utf8Path>>(&self, path: P) -> io::Result<()> {
         let path = from_utf8(path)?;
         self.cap_std.remove_file(path)
     }
@@ -289,7 +294,7 @@ impl Dir {
     /// This corresponds to [`std::fs::rename`], but only accesses paths
     /// relative to `self`.
     #[inline]
-    pub fn rename<P: AsRef<str>, Q: AsRef<str>>(
+    pub fn rename<P: AsRef<Utf8Path>, Q: AsRef<Utf8Path>>(
         &self,
         from: P,
         to_dir: &Self,
@@ -306,7 +311,11 @@ impl Dir {
     /// paths relative to `self`. Also, on some platforms, this function
     /// may fail if the file or directory cannot be opened for reading or
     /// writing first.
-    pub fn set_permissions<P: AsRef<str>>(&self, path: P, perm: Permissions) -> io::Result<()> {
+    pub fn set_permissions<P: AsRef<Utf8Path>>(
+        &self,
+        path: P,
+        perm: Permissions,
+    ) -> io::Result<()> {
         let path = from_utf8(path)?;
         self.cap_std.set_permissions(path, perm)
     }
@@ -316,7 +325,7 @@ impl Dir {
     /// This corresponds to [`std::fs::symlink_metadata`], but only accesses
     /// paths relative to `self`.
     #[inline]
-    pub fn symlink_metadata<P: AsRef<str>>(&self, path: P) -> io::Result<Metadata> {
+    pub fn symlink_metadata<P: AsRef<Utf8Path>>(&self, path: P) -> io::Result<Metadata> {
         let path = from_utf8(path)?;
         self.cap_std.symlink_metadata(path)
     }
@@ -326,7 +335,11 @@ impl Dir {
     /// This corresponds to [`std::fs::write`], but only accesses paths
     /// relative to `self`.
     #[inline]
-    pub fn write<P: AsRef<str>, C: AsRef<[u8]>>(&self, path: P, contents: C) -> io::Result<()> {
+    pub fn write<P: AsRef<Utf8Path>, C: AsRef<[u8]>>(
+        &self,
+        path: P,
+        contents: C,
+    ) -> io::Result<()> {
         let path = from_utf8(path)?;
         self.cap_std.write(path, contents)
     }
@@ -339,7 +352,11 @@ impl Dir {
     /// [`std::os::unix::fs::symlink`]: https://doc.rust-lang.org/std/os/unix/fs/fn.symlink.html
     #[cfg(not(windows))]
     #[inline]
-    pub fn symlink<P: AsRef<str>, Q: AsRef<str>>(&self, src: P, dst: Q) -> io::Result<()> {
+    pub fn symlink<P: AsRef<Utf8Path>, Q: AsRef<Utf8Path>>(
+        &self,
+        src: P,
+        dst: Q,
+    ) -> io::Result<()> {
         let src = from_utf8(src)?;
         let dst = from_utf8(dst)?;
         self.cap_std.symlink(src, dst)
@@ -353,7 +370,11 @@ impl Dir {
     /// [`std::os::windows::fs::symlink_file`]: https://doc.rust-lang.org/std/os/windows/fs/fn.symlink_file.html
     #[cfg(windows)]
     #[inline]
-    pub fn symlink_file<P: AsRef<str>, Q: AsRef<str>>(&self, src: P, dst: Q) -> io::Result<()> {
+    pub fn symlink_file<P: AsRef<Utf8Path>, Q: AsRef<Utf8Path>>(
+        &self,
+        src: P,
+        dst: Q,
+    ) -> io::Result<()> {
         let src = from_utf8(src)?;
         let dst = from_utf8(dst)?;
         self.cap_std.symlink_file(src, dst)
@@ -367,7 +388,11 @@ impl Dir {
     /// [`std::os::windows::fs::symlink_dir`]: https://doc.rust-lang.org/std/os/windows/fs/fn.symlink_dir.html
     #[cfg(windows)]
     #[inline]
-    pub fn symlink_dir<P: AsRef<str>, Q: AsRef<str>>(&self, src: P, dst: Q) -> io::Result<()> {
+    pub fn symlink_dir<P: AsRef<Utf8Path>, Q: AsRef<Utf8Path>>(
+        &self,
+        src: P,
+        dst: Q,
+    ) -> io::Result<()> {
         let src = from_utf8(src)?;
         let dst = from_utf8(dst)?;
         self.cap_std.symlink_dir(src, dst)
@@ -383,7 +408,7 @@ impl Dir {
     /// [`std::os::unix::net::UnixListener::bind`]: https://doc.rust-lang.org/std/os/unix/net/struct.UnixListener.html#method.bind
     #[cfg(unix)]
     #[inline]
-    pub fn bind_unix_listener<P: AsRef<str>>(&self, path: P) -> io::Result<UnixListener> {
+    pub fn bind_unix_listener<P: AsRef<Utf8Path>>(&self, path: P) -> io::Result<UnixListener> {
         let path = from_utf8(path)?;
         self.cap_std.bind_unix_listener(path)
     }
@@ -398,7 +423,7 @@ impl Dir {
     /// [`std::os::unix::net::UnixStream::connect`]: https://doc.rust-lang.org/std/os/unix/net/struct.UnixStream.html#method.connect
     #[cfg(unix)]
     #[inline]
-    pub fn connect_unix_stream<P: AsRef<str>>(&self, path: P) -> io::Result<UnixStream> {
+    pub fn connect_unix_stream<P: AsRef<Utf8Path>>(&self, path: P) -> io::Result<UnixStream> {
         let path = from_utf8(path)?;
         self.cap_std.connect_unix_stream(path)
     }
@@ -413,7 +438,7 @@ impl Dir {
     /// [`std::os::unix::net::UnixDatagram::bind`]: https://doc.rust-lang.org/std/os/unix/net/struct.UnixDatagram.html#method.bind
     #[cfg(unix)]
     #[inline]
-    pub fn bind_unix_datagram<P: AsRef<str>>(&self, path: P) -> io::Result<UnixDatagram> {
+    pub fn bind_unix_datagram<P: AsRef<Utf8Path>>(&self, path: P) -> io::Result<UnixDatagram> {
         let path = from_utf8(path)?;
         self.cap_std.bind_unix_datagram(path)
     }
@@ -428,7 +453,7 @@ impl Dir {
     /// [`std::os::unix::net::UnixDatagram::connect`]: https://doc.rust-lang.org/std/os/unix/net/struct.UnixDatagram.html#method.connect
     #[cfg(unix)]
     #[inline]
-    pub fn connect_unix_datagram<P: AsRef<str>>(
+    pub fn connect_unix_datagram<P: AsRef<Utf8Path>>(
         &self,
         unix_datagram: &UnixDatagram,
         path: P,
@@ -447,7 +472,7 @@ impl Dir {
     /// [`std::os::unix::net::UnixDatagram::send_to`]: https://doc.rust-lang.org/std/os/unix/net/struct.UnixDatagram.html#method.send_to
     #[cfg(unix)]
     #[inline]
-    pub fn send_to_unix_datagram_addr<P: AsRef<str>>(
+    pub fn send_to_unix_datagram_addr<P: AsRef<Utf8Path>>(
         &self,
         unix_datagram: &UnixDatagram,
         buf: &[u8],
@@ -472,7 +497,7 @@ impl Dir {
     /// This corresponds to [`std::path::Path::exists`], but only
     /// accesses paths relative to `self`.
     #[inline]
-    pub fn exists<P: AsRef<str>>(&self, path: P) -> bool {
+    pub fn exists<P: AsRef<Utf8Path>>(&self, path: P) -> bool {
         match from_utf8(path) {
             Ok(path) => self.cap_std.exists(path),
             Err(_) => false,
@@ -485,7 +510,7 @@ impl Dir {
     /// This corresponds to [`std::path::Path::is_file`], but only
     /// accesses paths relative to `self`.
     #[inline]
-    pub fn is_file<P: AsRef<str>>(&self, path: P) -> bool {
+    pub fn is_file<P: AsRef<Utf8Path>>(&self, path: P) -> bool {
         match from_utf8(path) {
             Ok(path) => self.cap_std.is_file(path),
             Err(_) => false,
@@ -499,7 +524,7 @@ impl Dir {
     /// traverse symbolic links to query information about the destination
     /// file. In case of broken symbolic links, this will return `false`.
     #[inline]
-    pub fn is_dir<P: AsRef<str>>(&self, path: P) -> bool {
+    pub fn is_dir<P: AsRef<Utf8Path>>(&self, path: P) -> bool {
         match from_utf8(path) {
             Ok(path) => self.cap_std.is_dir(path),
             Err(_) => false,
@@ -514,13 +539,45 @@ impl Dir {
     /// This function is not sandboxed and may access any path that the host
     /// process has access to.
     #[inline]
-    pub fn open_ambient_dir<P: AsRef<str>>(
+    pub fn open_ambient_dir<P: AsRef<Utf8Path>>(
         path: P,
         ambient_authority: AmbientAuthority,
     ) -> io::Result<Self> {
         let path = from_utf8(path)?;
         crate::fs::Dir::open_ambient_dir(path, ambient_authority).map(Self::from_cap_std)
     }
+
+    /// Constructs a new instance of `Self` by opening the parent directory
+    /// (aka "..") of `self`, using the host process' ambient authority.
+    ///
+    /// # Ambient Authority
+    ///
+    /// This function accesses a directory outside of the `self` subtree.
+    #[inline]
+    pub fn open_parent_dir<P: AsRef<Utf8Path>>(
+        &self,
+        ambient_authority: AmbientAuthority,
+    ) -> io::Result<Self> {
+        self.cap_std
+            .open_parent_dir(ambient_authority)
+            .map(Self::from_cap_std)
+    }
+
+    /// Recursively create a directory and all of its parent components if they
+    /// are missing, using the host process' ambient authority.
+    ///
+    /// # Ambient Authority
+    ///
+    /// This function is not sandboxed and may access any path that the host
+    /// process has access to.
+    #[inline]
+    pub fn create_ambient_dir_all<P: AsRef<Utf8Path>>(
+        path: P,
+        _ambient_authority: AmbientAuthority,
+    ) -> io::Result<()> {
+        let path = from_utf8(path)?;
+        fs::create_dir_all(path)
+    }
 }
 
 #[cfg(not(windows))]
